import AlteryxPythonSDK as Sdk
import connectors_core.alteryx.helpers as alteryx_helpers

import google.cloud.bigquery

import datetime
import decimal
import base64

class SchemaConversionException(Exception):
    def __init__(self, message):
        self.message = message

class BigQuerySchema:
    def __init__(self, table: google.cloud.bigquery.table.Table):
        self.table = table

    def get_record_info(self) -> Sdk.RecordInfo:
        """
        Wraps get_sdk_schema with alteryx_helpers.build_record_info
        :return: An instance of Sdk.RecordInfo for flattened representation of BigQuery table schema
        """
        return alteryx_helpers.build_record_info(self.alteryx_engine, self.get_sdk_schema())

    def get_sdk_schema(self, as_re=False):
        """
        Builds a list of dictionaries specifying the Alteryx SDK flattened representation of the BigQuery table schema
        :return: A list of dictionaries containing (description, name, Sdk.FieldType, size, scale) keys for each flattened field
        """
        sdk_schema = []
        for field in self.table.schema:
            self._get_sdk_schema_impl(sdk_schema, field, "")
        return sdk_schema

    def _get_sdk_schema_impl(self, sdk_schema, field, root_field_name):
        field_name = field.name if not root_field_name else root_field_name + '.' + field.name
        if field.mode == 'REPEATED':
            if field.field_type == 'RECORD':
                raise SchemaConversionException(
                    'Field {} has repeated record data type, which is not supported in Alteryx'.format(field_name)
                )
            sdk_schema.append({
                'description': field.description,
                'name': field_name,
                **self._get_sdk_type('STRING')
            })
        else:
            if field.field_type == 'RECORD':
                for subfield in field.fields:
                    self._get_sdk_schema_impl(sdk_schema, subfield, field_name)
            else:
                sdk_schema.append({
                    'description': field.description,
                    'name': field_name,
                    **self._get_sdk_type(field.field_type)
                })

    def _get_sdk_type(self, cloud_type):
        """
        Helper for converting Big Query types to Alteryx SDK Types
        :param cloud_type: str representation of Big Query Google Cloud Type
        :return: dictionary representation of Alteryx SDK Type
        """
        try:
            return {
                'STRING': {'type': Sdk.FieldType.v_wstring, 'size': self.max_field_length, 'scale': 0},
                'BYTES': {'type': Sdk.FieldType.blob, 'size': self.max_field_length, 'scale': 0},
                'INTEGER': {'type': Sdk.FieldType.int64, 'size': 0, 'scale': 0},
                'NUMERIC': {'type': Sdk.FieldType.fixeddecimal, 'size': 38, 'scale': 9}, # see https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#numeric-type
                'FLOAT': {'type': Sdk.FieldType.float, 'size': 0, 'scale': 0},
                'BOOLEAN': {'type': Sdk.FieldType.bool, 'size': 0, 'scale': 0},
                'TIMESTAMP': {'type': Sdk.FieldType.datetime, 'size': 0, 'scale': 0},
                'TIME': {'type': Sdk.FieldType.time, 'size': 0, 'scale': 0},
                'DATE': {'type': Sdk.FieldType.date, 'size': 0, 'scale': 0},
                'DATETIME': {'type': Sdk.FieldType.datetime, 'size': 0, 'scale': 0},
                'RECORD': {'type': Sdk.FieldType.blob, 'size': self.max_field_length, 'scale': 0},
            }[cloud_type]
        except KeyError as e:
            return {'type': Sdk.FieldType.wstring, 'size': self.max_field_length, 'scale': 0}

class BigQuerySchemaWriter(BigQuerySchema):
    def __init__(self, table, record_info: Sdk.RecordInfo):
        super().__init__(table)
        self.record_info = record_info
        self.array_types_dict = {}
        self.schema_checked = False

    def _get_fields(self, field_info):
        if hasattr(field_info, 'fields'):
            return field_info.fields
        if hasattr(field_info, 'schema'):
            return field_info.schema
        return []

    def _is_expected(self, field_name, field_info):
        name_components = field_name.split('.')
        array_field_type = None
        for name in name_components:
            is_valid_name = False
            for subfield in self._get_fields(field_info):
                if subfield.name == name:
                    field_info = subfield
                    is_valid_name = True
                    if subfield.mode == 'REPEATED':
                        array_field_type = subfield.field_type
                    break
            if not is_valid_name:
                return False
        # Cache field types in order to know array data types
        if array_field_type:
            self.array_types_dict[field_name] = array_field_type
        return True

    def build_row_from_record(self, record_ref, alteryx_engine: Sdk.AlteryxEngine = None):
        my_data = {}
        try:
            for i in range(0, self.record_info.num_fields):
                field = self.record_info[i]  # type: Sdk.Field
                if not self.schema_checked:
                    if not self._is_expected(field.name, self.table):
                        if alteryx_engine:
                            raise SchemaConversionException(alteryx_engine.xmsg(
                                'Unexpected field name for destination table schema: @1', field.name
                            ))
                        else:
                            raise SchemaConversionException(
                                'Unexpected field name for destination table schema: {}'.format(field.name)
                            )
                field_name_components = field.name.split('.')
                field_val = alteryx_helpers.get_value_for_field(field, record_ref)
                self._build_field(my_data, field_name_components, field_val, field.type, self.array_types_dict.get(field.name))
            self.schema_checked = True
        except Exception as e:
            error_message = e.message if hasattr(e, "message") else ''
            if alteryx_engine:
                raise SchemaConversionException(alteryx_engine.xmsg(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record. @1', error_message
                ))
            else:
                raise SchemaConversionException(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record. {}'.format(error_message)
                )
        return my_data

    def _convert_to_type(self, value, field_type):
        if field_type == 'INTEGER':
            return int(value)
        elif field_type == 'FLOAT':
            return float(value)
        elif field_type == 'BOOLEAN':
            return bool(value)
        elif field_type == 'NUMERIC':
            return decimal.Decimal(value)
        elif field_type == 'STRING':
            return value[1:-1] # Trim last ' characters
        return value

    def _build_field(self, payload, field_name_components, field_val, field_type, array_field_type):
        top_attr = field_name_components[0]

        if len(field_name_components) == 1:
            if type(field_val) is bytearray:
                field_val = base64.standard_b64encode(field_val).decode("ascii")
            if type(field_val) is datetime.timedelta:
                field_val = str(field_val)
            if field_type == Sdk.FieldType.date and field_val:
                field_val = str(field_val).split(' ')[0]

            if not array_field_type:
                payload[top_attr] = field_val
            else:
                payload[top_attr] = []
                if field_val:
                    array_items = field_val[1:-1].split(',') # We expect array values between braces like "[5,4,-8,7]"
                    for item in array_items:
                        payload[top_attr].append(self._convert_to_type(item.strip(), array_field_type))
        else:
            if top_attr not in payload.keys():
                payload[top_attr] = {}
            self._build_field(payload[top_attr], field_name_components[1:], field_val, field_type, array_field_type)

class BigQuerySchemaReader(BigQuerySchema):
    def __init__(self,
                 alteryx_engine: Sdk.AlteryxEngine,
                 table: google.cloud.bigquery.table.Table,
                 max_field_length: int):
        super().__init__(table)
        self.alteryx_engine = alteryx_engine
        self.max_field_length = max_field_length

    def _get_field_value(self, data, field_name):
        field_value = None
        if not isinstance(data, dict):
            if hasattr(data, field_name):
                field_value = getattr(data, field_name)
        else:
            if field_name in data:
                field_value = data[field_name]
        return field_value

    def generate_records_from_field(self, result, field, record_info_out, root_field_name):
        """
        Yields as many (record_field, field_value) pairs as necessary for the given field

        :param result: An instance of google.cloud.bigquery.table.Row; provides field values
        :param field: An instance of google.cloud.bigquery.schema.SchemaField; the field to generate a record for
        :param record_info_out: An instance of Sdk.RecordInfo
        :return: (record_field, field_value) pairs
        """
        field_name = field.name if not root_field_name else root_field_name + '.' + field.name
        if field.mode == 'REPEATED':
            if field.field_type == 'RECORD':
                raise SchemaConversionException(
                    'Field {} has repeated record data type, which is not supported in Alteryx'.format(field_name)
                )
            array_field_value = self._get_field_value(result, field.name)
            if not array_field_value:
                array_field_value = None
            idx = record_info_out.get_field_num(field_name)
            record_field = record_info_out[idx]
            yield record_field, str(array_field_value) if array_field_value else array_field_value
        else:
            if field.field_type == 'RECORD':
                for subfield in field.fields:
                    yield from self.generate_records_from_field(
                        self._get_field_value(result, field.name), subfield, record_info_out, field_name
                    )
            else:
                idx = record_info_out.get_field_num(field_name)
                record_field = record_info_out[idx]
                yield record_field, self._get_field_value(result, field.name)
