import AlteryxPythonSDK as Sdk
from requests_oauthlib import OAuth2Session
from connectors_core.oauth2.clients import AzureActiveDirectoryWebApplicationClient
import requests
import functools
from datetime import datetime
import json

from requests.exceptions import ConnectionError
from urllib3.exceptions import MaxRetryError, NewConnectionError
from socket import gaierror
from oauthlib.oauth2.rfc6749.errors import InvalidClientIdError, TokenExpiredError

import connectors_core.alteryx.helpers as alteryx_helpers
from connectors_core import alteryx
from .common import (
    FailToAuthenticateOnTokenExpiration,
    OUTPUT_MODE_REPLACE_DATASET_IF_EXISTS,
    OUTPUT_MODE_RENAME_DATASET_IF_EXISTS,
    OUTPUT_MODE_APPEND_TO_TABLE_IF_EXISTS,
    AUTHORIZATION_SERVER_URL,
    BASE_API_ENDPOINT,
    get_power_bi_type,
    get_table_schema,
    get_api_url,
    serialize_datetime,
    RESOURCE_URI
)

class PowerBIOutputConnections(alteryx.IncomingConnectionBase):
    def __init__(self, parent: alteryx.AyxPluginBase):
        super().__init__(parent)
        self.record_info = None  # type: Sdk.RecordInfo
        self.is_initialized = False
        self.row_buffer = []
        self.row_insert_batch_size = None
        self.tenant_id = None  # type: str
        self.client_id = None  # type: str
        self.client_secret = None  # type: str
        self.token = None  # type: dict
        self.workspace_id = None    # type: str
        self.dataset_name = None    # type: str
        self.table_name = None    # type: str
        self.output_mode = None # type: str
        self.has_custom_dataset_table_names = None    # type: bool
        self.dataset_id = None    # type: str
        self.table_id = None    # type: str

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        self.dataset_is_initialized = False
        if not self.parent.is_initialized:
            return False
        self.record_info = record_info_in
        self.tenant_id = self.parent.tenant_id
        self.client_id = self.parent.client_id
        self.client_secret = self.parent.client_secret
        self.token = self.parent.token
        self.workspace_id = self.parent.workspace_id
        self.dataset_name = self.parent.dataset_name
        self.table_name = self.parent.table_name
        self.output_mode = self.parent.output_mode
        self.row_insert_batch_size = self.parent.row_insert_batch_size
        self.has_custom_dataset_table_names = self.parent.has_custom_dataset_table_names
        self.dataset_id = self.parent.dataset_id
        self.table_id = self.parent.table_id
        self.power_bi_schema = get_table_schema(
            self.dataset_name,
            self.table_name,
            self.record_info
        )
        client = AzureActiveDirectoryWebApplicationClient(
            client_id=self.client_id,
            client_secret=self.client_secret,
            token=self.token
        )
        self.authority_uri = '{}/{}/oauth2/token'.format(AUTHORIZATION_SERVER_URL, self.tenant_id)
        refresh_kwargs = {}
        self.session = OAuth2Session(
            client_id=self.client_id,
            client=client,
            token=self.token,
            auto_refresh_url=self.authority_uri,
            auto_refresh_kwargs=refresh_kwargs,
            token_updater=self._update_token
        )
        return True

    def _update_token(self, token: dict):
        self.token = token
        self.log.debug("Token updated: expires_at={}".format(token['expires_at']))

    def catch_token_expiration(request_fn):
        """
        A function decorator to try/except TokenExpiredErrors and attempt to
        re-authenticate using client credentials instead of refresh token
        """
        @functools.wraps(request_fn)
        def wrapped_fn(self, *args, **kwargs):
            try:
                # Make request
                self.log.debug('Making initial request from catch_token_expiration [{}]'.format(request_fn.__name__))
                return request_fn(self, *args, **kwargs)
            except (ConnectionError, MaxRetryError, NewConnectionError, gaierror) as error:
                self.log.exception('Could not connect to PowerBI on initial request')
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    'Could not connect to Power BI service. Please check your network connectivity.'
                ))
                raise
            except (TokenExpiredError, InvalidClientIdError, Exception) as e:
                # OAuth2Session failed to refresh automatically with refresh_token
                # Try to refresh from user credentials if they were provided
                if not (self.parent.username and self.parent.password):
                    self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                        'Unable to refresh access token. Username and password credentials not persisted.',
                    ))
                    raise FailToAuthenticateOnTokenExpiration(self.parent.alteryx_engine.xmsg(
                        'Unable to refresh access token. Username and password credentials not persisted.'
                    ))
                self.log.debug('TokenExpiredError caught, attempting to refresh using client credentials')
                payload_dict = {
                    'client_id': self.parent.client_id,
                    'client_secret': self.parent.client_secret,
                    'username': self.parent.username,
                    'password': self.parent.password,
                    'grant_type': 'password',
                    'resource': RESOURCE_URI
                }
                response = requests.post(
                    self.authority_uri,
                    data=payload_dict
                )
                if response.ok:
                    self.token = response.json()
                    self.session.token = self.token
                    self.log.debug('Making request from catch_token_expiration with refreshed token [{}]'.format(request_fn.__name__))
                    return request_fn(self, *args, **kwargs)
                else:
                    self.log.error(
                        'There was an error refreshing access token using end user credentials: {} {}'.format(response.status_code, response.content)
                    )
                    self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                        'Unable to refresh access token from persisted username and password credentials.'
                    ))
                    raise FailToAuthenticateOnTokenExpiration(self.parent.alteryx_engine.xmsg(
                        'Unable to refresh access token from persisted username and password credentials.'
                    ))
        return wrapped_fn

    @property
    def headers(self):
        if not self.token:
            raise ValueError('No token available to create headers')
        return {
            'Authorization': 'Bearer {}'.format(self.token['access_token']),
            'Content-Type': 'application/json'
        }

    @property
    def datasets_endpoint(self):
        if self.workspace_id != 'MyWorkspace':
            return f'groups/{self.workspace_id}/datasets'
        return 'datasets'

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        if not self.dataset_is_initialized:
            if not self.try_init_dataset():
                return False
        record_payload = {}
        for i in range(0, self.record_info.num_fields):
            field = self.record_info[i]  # type: Sdk.Field
            try:
                record_payload[field.name] = alteryx_helpers.get_value_for_field(field, in_record)
            except KeyError as e:
                self.log.exception('Could not get value for field {}'.format(field.name))
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    'Could not get value for field @1: @2', field.name, str(e)
                ))
                return False
        self.log.debug('Adding record to row_buffer: {}'.format(record_payload))
        self.row_buffer.append(record_payload)
        if len(self.row_buffer) >= self.row_insert_batch_size:
            return self.post_rows()
        return True

    def ii_update_progress(self, d_percent: float):
        try:
            self.parent.alteryx_engine.output_tool_progress(
                self.parent.n_tool_id, d_percent
            )
        except AttributeError:
            pass

    def ii_close(self):
        if self.row_buffer:
            success = self.post_rows()
            if not success:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    'Closing connection to Power BI. Failed to post all rows.'
                ))
        self.is_initialized = False

    @catch_token_expiration
    def post_rows(self):
        buffer_slice = self.row_buffer[:self.row_insert_batch_size]
        url = get_api_url('{}/{}/tables/{}/rows'.format(
                self.datasets_endpoint,
                self.dataset_id,
                self.table_name
            ))
        try:
            json_rows = json.dumps({"rows": buffer_slice}, default=serialize_datetime)
        except TypeError as e:
            self.log.exception('JSON Serialization TypeError in [post_rows]')
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'JSON Serialization Error. Failed to push record(s) to Power BI. @1',
                str(e)
            ))
            return False
        self.log.debug('POST [post_rows] {} with body: {}'.format(url, json_rows))
        response = self.session.post(url, json_rows, headers=self.headers)
        if not response.ok:
            try:
                error = response.json()['error']['message']
            except (ValueError, KeyError) as e:
                error = self.parent.alteryx_engine.xmsg(
                    'There was an unexpected error'
                )
            self.log.error('Error in [ii_push_record]: {}'.format(error))
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'Failed to push record(s) to Power BI. @1: @2',
                str(response.status_code), error
            ))
            return False
        self.row_buffer = self.row_buffer[self.row_insert_batch_size:]
        return True

    @catch_token_expiration
    def table_exists_in_dataset(self, dataset_id, table_name):
        if not dataset_id:
            return False
        url = get_api_url(f'{self.datasets_endpoint}/{dataset_id}/tables')
        self.log.debug('GET [table_exists_in_dataset] {}'.format(url))
        response = self.session.get(url)
        if response.ok:
            for table in response.json()['value']:
                if table['name'] == table_name:
                    return True
        else:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'Failed to check if dataset @1 exists. @2: @3',
                self.dataset_name, str(response.status_code), response.reason
            ))
        return False

    @catch_token_expiration
    def get_datasets(self):
        url = get_api_url(self.datasets_endpoint)
        self.log.debug('GET [get_datasets] {}'.format(url))
        try:
            response = self.session.get(url)
            if response.ok:
                return response.json()['value']
            else:
                self.log.error('[get_datasets] received response from Microsoft: {}'.format(response.status_code))
        except (KeyError, ValueError) as e:
            self.log.error('Failed to get value key in JSON response')
        self.parent.display_error_msg('Failed to determine if dataset already exists.')
        return []

    def get_next_unique_dataset_name(self, dataset_name):
        dataset_names = {dataset['name'] for dataset in self.get_datasets()}
        name = dataset_name
        counter = 0
        while name in dataset_names:
            name = '{}_{}'.format(dataset_name, counter)
            counter += 1
        return name

    def get_dataset_id_from_name(self, dataset_name):
        for dataset in self.get_datasets():
            if dataset['name'] == dataset_name:
                return dataset['id']
        return None
    
    def get_dataset_name_from_id(self, dataset_id):
        for dataset in self.get_datasets():
            if dataset['id'] == dataset_id:
                return dataset['name']
        return None

    def append_to_table_init(self):
        table_exists = self.table_exists_in_dataset(
            self.dataset_id, self.table_name
        )
        self.log.debug('Table Exists? {}'.format(table_exists))
        if not table_exists:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'Dataset @1 exists without table @2. Aborting workflow',
                self.dataset_name,
                self.table_name
            ))
            # Cannot write to the dataset because it does not contain the table
            self.dataset_id = None

    def replace_dataset_init(self):
        self.parent.display_info_msg(self.parent.alteryx_engine.xmsg(
            'Dataset @1 existed. Overwriting the existing data',
            self.dataset_name
        ))
        self.delete_dataset(self.dataset_id)
        self.dataset_id = self.create_dataset(
            self.dataset_name, self.power_bi_schema
        )

    def rename_dataset_init(self):
        new_dataset_name = self.get_next_unique_dataset_name(
            self.dataset_name
        )
        self.parent.display_info_msg(self.parent.alteryx_engine.xmsg(
            'Dataset @1 existed. Creating Dataset with new name @2',
            self.dataset_name,
            new_dataset_name
        ))
        self.dataset_name = new_dataset_name
        self.dataset_id = self.create_dataset(
            self.dataset_name, self.power_bi_schema
        )

    def try_init_dataset(self):
        # Check if dataset already exists
        self.dataset_id = self.get_dataset_id_from_name(self.dataset_name) if self.has_custom_dataset_table_names else self.dataset_id
        self.dataset_name = self.dataset_name if self.has_custom_dataset_table_names else self.get_dataset_name_from_id(self.dataset_id)
        self.table_name = self.table_name if self.has_custom_dataset_table_names else self.table_id
        self.log.debug('Dataset Exists? {}'.format(self.dataset_id is not None))
        if not self.dataset_id:
            # Always create dataset if doesn't exist, requires creating a table
            self.dataset_id = self.create_dataset(
                self.dataset_name, self.power_bi_schema
            )
        else:
            # Dataset already existed
            if self.output_mode == OUTPUT_MODE_APPEND_TO_TABLE_IF_EXISTS:
                self.append_to_table_init()
            elif self.output_mode == OUTPUT_MODE_REPLACE_DATASET_IF_EXISTS:
                self.replace_dataset_init()
            elif self.output_mode == OUTPUT_MODE_RENAME_DATASET_IF_EXISTS:
                self.rename_dataset_init()
        if not self.dataset_id:
            return False
        self.dataset_is_initialized = True
        return True

    @catch_token_expiration
    def create_dataset(self, dataset_name, tables_schema):
        payload = json.dumps({
            "name": dataset_name,
            "tables": [ tables_schema ]
        })
        url = get_api_url(self.datasets_endpoint)
        self.log.debug('POST [create_dataset] {} with body {}'.format(url, payload))
        response = self.session.post(url, payload, headers=self.headers)
        if response.ok:
            new_dataset_id = response.json()['id']    # New Dataset's ID
            return new_dataset_id
        else:
            self.log.error('Error creating a dataset {}'.format(response.content))
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'Failed to create dataset @1. @2: @3',
                self.dataset_name, str(response.status_code), response.reason
            ))

    @catch_token_expiration
    def delete_dataset(self, dataset_id):
        url = get_api_url(f'{self.datasets_endpoint}/{self.dataset_id}')
        self.log.debug('DELETE [delete_dataset] {}'.format(url))
        response = self.session.delete(url, headers=self.headers)
        self.dataset_id = None
        if not response.ok:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                'Failed to delete existing dataset @1. @2: @3',
                self.dataset_name, str(response.status_code), response.reason
            ))
