import json
import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx

from typing import Union

from .connections import PowerBIOutputConnections
from .common import OUTPUT_MODES, BASE_API_ENDPOINT

class PowerBIOutputEngine(alteryx.AyxPluginBase):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.tenant_id = None  # type: str
        self.client_id = None  # type: str
        self.client_secret = None  # type: str
        self.username = None
        self.password = None
        self.token = None  # type: dict
        self.output_mode = None  # type: str
        self.incoming_connection_added = False
        self.workspace_id = None    # type: str
        self.dataset_name = None    # type: str
        self.table_name = None    # type: str
        self.has_custom_dataset_table_names = None    # type: bool
        self.dataset_id = None    # type: str
        self.table_id = None    # type: str
        self.row_insert_batch_size = None   # type: int

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg(self.alteryx_engine.xmsg(
                "Only one incoming connection can be accepted by this tool"
            ))
            return None
        connection_handler = PowerBIOutputConnections(self)
        return connection_handler

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)
        self.tenant_id = self.get_xml_config_value(str_xml, 'tenantId')
        self.client_id = self.get_xml_config_value(str_xml, 'clientId')
        self.client_secret = self.get_encrypted_xml_config_value(str_xml, 'clientSecret')
        self.username = self.get_xml_config_value(str_xml, 'username')
        self.password = self.get_encrypted_xml_config_value(str_xml, 'password')
        self.output_mode = self.get_xml_config_value(str_xml, 'outputMode')
        self.workspace_id = self.get_xml_config_value(str_xml, 'workspace')
        self.dataset_name = self.get_xml_config_value(str_xml, 'datasetName')
        self.table_name = self.get_xml_config_value(str_xml, 'tableName')
      
        self.has_custom_dataset_table_names = True if self.get_xml_config_value(str_xml, 'enterCustomDatasetAndTable') == 'True' else False
        self.dataset_id = self.get_xml_config_value(str_xml, 'datasetId')
        self.table_id = self.get_xml_config_value(str_xml, 'tableId')
        try:
            self.row_insert_batch_size = int(self.get_xml_config_value(str_xml, 'rowInsertBatchSize'))
        except (ValueError, TypeError) as e:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Please enter a valid row insert batch size'
            ))

        if not self.client_id:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Please enter a client ID'
            ))
        if not self.client_secret:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Please enter a client secret'
            ))
        if not self.output_mode:
            self.invalidate_config(self.alteryx_engine.xmsg(
                "Please select an output operation"
            ))
        if self.output_mode not in OUTPUT_MODES:
            self.invalidate_config(self.alteryx_engine.xmsg(
                "Please select a valid output operation."
            ))
        if not self.workspace_id:
            self.invalidate_config(self.alteryx_engine.xmsg(
                "Please select a workspace"
            ))
        if self.has_custom_dataset_table_names:
            if not self.dataset_name:
                self.invalidate_config(self.alteryx_engine.xmsg(
                    "Please enter a dataset name"
                ))
            if not self.table_name:
                self.invalidate_config(self.alteryx_engine.xmsg(
                    "Please enter a table name"
                ))
        else:
            if not self.dataset_id:
                self.invalidate_config(self.alteryx_engine.xmsg(
                    "Please select a dataset"
                ))
            if not self.table_id:
                self.invalidate_config(self.alteryx_engine.xmsg(
                    "Please select a table"
                ))

        token_str = self.get_encrypted_xml_config_value(str_xml, 'accessToken')
        if not token_str:
            self.invalidate_config(self.alteryx_engine.xmsg('Please log in'))
        elif self.is_initialized:
            try:
                self.token = json.loads(token_str)
                required_fields = ['expires_at', 'resource', 'access_token', 'refresh_token']
                if any(req not in self.token for req in required_fields):
                    self.invalidate_config(self.alteryx_engine.xmsg(
                        'Invalid auth token. Missing one or more required fields'
                    ))
            except Exception as e:
                self.invalidate_config(self.alteryx_engine.xmsg(
                    "Failed parsing authentication: @1", str(e)
                ))

        metadata_dict = {
            'Host': BASE_API_ENDPOINT,
            'ServiceName': 'PowerBI',
            'PluginName': 'PowerBI',
            'Database': self.dataset_name,
            'Table': self.table_name
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor


    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg(self.alteryx_engine.xmsg(
            'This tool does not allow outgoing connections'
        ))
        return False

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()
